import logging

from django.core.exceptions import ImproperlyConfigured
from django.apps import apps
from celery import shared_task
from haystack.exceptions import NotHandled as IndexNotFoundException
from haystack import connections, connection_router


logger = logging.getLogger(__name__)


def split_identifier(identifier):
    """
    Break down the identifier representing the instance.
    Converts 'notes.note.23' into ('notes.note', 23).
    """
    bits = identifier.split('.')

    if len(bits) < 2:
        logger.error("Unable to parse object "
                     "identifer '%s'. Moving on..." % identifier)
        return (None, None)

    pk = bits[-1]
    # In case Django ever handles full paths...
    object_path = '.'.join(bits[:-1])
    return (object_path, pk)


def get_instance(model_class, pk):
    """
    Fetch the instance in a standarized way.
    """
    instance = None
    try:
        instance = model_class._default_manager.get(pk=pk)
    except model_class.DoesNotExist:
        logger.error("Couldn't load %s.%s.%s. Somehow it went missing?" %
                     (model_class._meta.app_label.lower(),
                      model_class._meta.object_name.lower(), pk))
    except model_class.MultipleObjectsReturned:
        logger.error("More than one object with pk %s. Oops?" % pk)
    return instance


def get_indexes(model_class):
    """
    Fetch the model's registered ``SearchIndex`` in a standarized way.
    """
    try:
        using_backends = connection_router.for_write(
            models=[model_class]
        )
        for using in using_backends:
            index_holder = connections[using].get_unified_index()
            yield index_holder.get_index(model_class), using
    except IndexNotFoundException:
        raise ImproperlyConfigured(
            "Couldn't find a SearchIndex for %s." % model_class
        )


def get_model_class(object_path):
    """
    Fetch the model's class in a standarized way.
    """
    bits = object_path.split('.')
    app_name = '.'.join(bits[:-1])
    classname = bits[-1]
    model_class = apps.get_model(app_name, classname)

    if model_class is None:
        raise ImproperlyConfigured("Could not load model '%s'." %
                                   object_path)
    return model_class


@shared_task
def update_index(
    action,
    identifier,
):
    logger.debug("Update Index")
    object_path, pk = split_identifier(identifier)
    if object_path is None or pk is None:
        msg = "Couldn't handle object with identifier %s" % identifier
        logger.error(msg)
        raise ValueError(msg)

    # Then get the model class for the object path
    model_class = get_model_class(object_path)
    logger.debug(f"model_class={model_class}")
    for current_index, using in get_indexes(model_class):
        current_index_name = ".".join([current_index.__class__.__module__,
                                       current_index.__class__.__name__])

        if action == 'delete':
            # If the object is gone, we'll use just the identifier
            # against the index.
            try:
                current_index.remove_object(identifier, using=using)
            except Exception as exc:
                logger.exception(exc)
            else:
                msg = ("Deleted '%s' (with %s)" %
                       (identifier, current_index_name))
                logger.debug(msg)
        elif action == 'save':
            # and the instance of the model class with the pk
            instance = get_instance(model_class, pk)
            if instance is None:
                logger.debug("Failed updating '%s' (with %s)" %
                             (identifier, current_index_name))
                raise ValueError("Couldn't load object '%s'" % identifier)

            # Call the appropriate handler of the current index and
            # handle exception if necessary
            try:
                current_index.update_object(instance, using=using)
            except Exception as exc:
                logger.exception(exc)
            else:
                msg = ("Updated '%s' (with %s)" %
                       (identifier, current_index_name))
                logger.debug(msg)
        else:
            logger.error("Unrecognized action '%s'. Moving on..." % action)
            raise ValueError("Unrecognized action %s" % action)



import uuid

from django.contrib.auth.models import AbstractUser, Permission
from django.db import models

from papermerge.core.models.document import Document
from papermerge.core.models.folder import Folder
from papermerge.core.models.node import BaseTreeNode, AbstractNode
from papermerge.core.models.page import Page
from papermerge.core.models.tags import (
    ColoredTag,
    Tag
)

from papermerge.core.models.document_version import DocumentVersion


class User(AbstractUser):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)

    # Initially user is created with empty `home_folder` and `inbox_folder`.
    # Home and Inbox folder fields are populated as part of `post_save` model
    # `user` signal
    home_folder = models.OneToOneField(
        'Folder',
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        related_name='home_folder_of'
    )
    inbox_folder = models.OneToOneField(
        'Folder',
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        related_name='inbox_folder_of'
    )

    created_at = models.DateTimeField(
        auto_now_add=True
    )
    updated_at = models.DateTimeField(
        auto_now=True
    )

    def create_special_folders(self):
        """
        Creates user's home and inbox folders.

        This method is invoked in user's post save
        signal on model creation.
        """
        _inbox, _ = Folder.objects.get_or_create(
            title=Folder.INBOX_TITLE,
            parent=None,
            user=self
        )
        _home, _ = Folder.objects.get_or_create(
            title=Folder.HOME_TITLE,
            parent=None,
            user=self
        )
        self.inbox_folder = _inbox
        self.home_folder = _home
        self.save()

    @property
    def perm_codenames(self):
        """
        Returns aggregated list of permissions codenames of the user.

        Permissions are aggregated from self.user_permissions +
        permissions from all self.groups.
        In Django permissions can be assigned directly to user object
        or via groups. This attribute returns list of string codenames
        of all permissions - from user object and from each associated
        group. It is meant to be passed (and used by) the frontend
        in order to toggle on/off different parts depending on what
        user is allowed/not allowed to view/perform.

        Note that codename DOES NOT include app label.
        """

        # 1. gather all perms via associated groups
        user_groups_field = self._meta.get_field('groups')
        user_groups_query = 'group__%s' % user_groups_field.related_query_name()
        result1 = Permission.objects.filter(
            **{user_groups_query: self}
        ).values_list(
            'codename',
            flat=True
        )
        # 2. gather all perms via user obj
        result2 = self.user_permissions.values_list(
            'codename',
            flat=True
        )

        # combine 1. and 2.
        result = list(result1) + list(result2)

        # remove all duplicates
        return list(set(result))

    def delete(self, using=None, keep_parents=False):
        Document.objects.filter(user=self).delete()
        super().delete(using=using, keep_parents=keep_parents)



import pytz
import uuid

from django.utils import timezone
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.contrib.contenttypes.models import ContentType

from taggit.managers import TaggableManager

from papermerge.core import validators
from papermerge.core.models.tags import ColoredTag
from polymorphic_tree.models import (
    PolymorphicMPTTModel,
    PolymorphicTreeForeignKey
)
from polymorphic_tree.managers import (
    PolymorphicMPTTModelManager,
    PolymorphicMPTTQuerySet
)

# things you can propagate from parent node to
# child node
PROPAGATE_ACCESS = 'access'
PROPAGATE_KV = 'kv'
PROPAGATE_KVCOMP = 'kvcomp'
RELATED_NAME_FMT = "%(app_label)s_%(class)s_related"
RELATED_QUERY_NAME_FMT = "%(app_label)s_%(class)ss"


class NodeManager(PolymorphicMPTTModelManager):
    pass


class NodeQuerySet(PolymorphicMPTTQuerySet):

    def delete(self, *args, **kwargs):
        for node in self:
            descendants = node.get_descendants()

            if descendants.count() > 0:
                descendants.delete(*args, **kwargs)
            # At this point all descendants were deleted.
            # Self delete :)
            try:
                node.delete(*args, **kwargs)
            except BaseTreeNode.DoesNotExist:
                # this node was deleted by earlier recursive call
                # it is ok, just skip
                pass


CustomNodeManager = NodeManager.from_queryset(NodeQuerySet)


class BaseTreeNode(PolymorphicMPTTModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4)

    parent = PolymorphicTreeForeignKey(
        'self',
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        related_name='children',
        verbose_name=_('parent')
    )
    title = models.CharField(
        _("Title"),
        max_length=200,
        validators=[validators.safe_character_validator]
    )

    lang = models.CharField(
        _('Language'),
        max_length=8,
        blank=False,
        null=False,
        default='deu'
    )

    user = models.ForeignKey(
        'User',
        on_delete=models.CASCADE
    )

    created_at = models.DateTimeField(
        auto_now_add=True
    )
    updated_at = models.DateTimeField(
        auto_now=True
    )

    tags = TaggableManager(through=ColoredTag)

    # custom Manager + custom QuerySet
    objects = CustomNodeManager()

    @property
    def idified_title(self):
        """
        Returns a title with ID part inserted at the end

        Example:
            input: title="My Invoices", id="233453"
            output: "My Invoices-233453"
        """

        return f'{self.title}-{self.id}'

    def human_datetime(self, _datetime) -> str:
        """
        Localize and format datetime instance considering user preferences.
        """
        tz = pytz.timezone(
            self.user.preferences['localization__timezone']
        )
        fmt = self.user.preferences['localization__date_format']
        fmt += " " + self.user.preferences['localization__time_format']

        ret_datetime = timezone.localtime(_datetime, timezone=tz)

        ret = ret_datetime.strftime(fmt)
        return ret

    @property
    def human_updated_at(self) -> str:
        """
        updated_at displayed considering user's timezone, date and time prefs.

        returns string with user friendly formated datetime.
        """

        ret = self.human_datetime(self.updated_at)
        return ret

    @property
    def human_created_at(self) -> str:
        """
        created_at displayed considering user's timezone, date and time prefs.
        """
        ret = self.human_datetime(self.created_at)
        return ret

    def is_folder(self):
        folder_ct = ContentType.objects.get(
            app_label='core', model='folder'
        )
        return self.polymorphic_ctype_id == folder_ct.id

    def is_document(self):
        document_ct = ContentType.objects.get(
            app_label='core', model='document'
        )
        return document_ct.id == self.polymorphic_ctype_id

    class Meta(PolymorphicMPTTModel.Meta):
        # please do not confuse this "Documents" verbose name
        # with real Document object, which is derived from BaseNodeTree.
        # The reason for this naming confusing is that from the point
        # of view of users, the BaseNodeTree are just a list of documents.
        verbose_name = _("Documents")
        verbose_name_plural = _("Documents")
        _icon_name = 'basetreenode'


class AbstractNode(models.Model):
    """
    Common class apps need to inherit from in order
    to extend BaseTreeNode model.
    """
    base_ptr = models.ForeignKey(
        BaseTreeNode,
        related_name=RELATED_NAME_FMT,
        related_query_name=RELATED_QUERY_NAME_FMT,
        on_delete=models.CASCADE
    )

    class Meta:
        abstract = True

    def get_title(self):
        return self.base_ptr.title
